=== modified file 'loggerhead.conf'
--- loggerhead.conf	2006-12-20 09:29:42 +0000
+++ loggerhead.conf	2006-12-20 21:17:53 +0000
@@ -3,23 +3,34 @@
 # (you only need this if you don't install bzr.)
 bzrpath='/Users/robey/code/bzr/bzr'
 
-# each branch is in a sub-folder of the website, and the config
-# key describes that folder name:
-[bazaar-dev]
-
-    # where is the bazaar branch located, and what should we call it?
-    folder='/Users/robey/code/bzr/bzr.dev'
-
-    # (optional) is there a friendier name for the branch that i should
-    # use on the web pages?
-    #branch_name='bazaar-dev'
-
-    # should we cache revisions?  (HIGHLY recommended)  where?
-    # this will be a folder containing a couple of files.  if the folder
-    # doesn't already exist, we'll create it.
-    cachepath='/Users/robey/code/bzr/bzr.dev/.bzr/loggerhead-cache'
-
-    # what is my externally-visible url? (for use in the atom feed)
-    external_url="http://localhost:8080"
-
+# if you want a special title on the front page, set it here:
+title='branches in loggerhead'
+
+# group branches by project:
+[bazaar]
+
+    # name of this project
+    name = 'bazaar'
+
+    # each branch is in a sub-folder of the website, and the config
+    # key describes that folder name:
+    [[bazaar-dev]]
+
+        # where is the bazaar branch located?
+        folder='/Users/robey/code/bzr/bzr.dev'
+
+        # (optional) is there a friendier name for the branch that i should
+        # use on the web pages?
+        #branch_name='bazaar-dev'
+
+        # (optional) short description for the top-level page
+        description="bazaar vcs development branch"
+
+        # (optional) public branch url for the top-level page
+        url='http://bazaar-ng.org/bzr/bzr.dev'
+
+        # should we cache revisions?  (HIGHLY recommended)  where?
+        # this will be a folder containing a couple of files.  if the folder
+        # doesn't already exist, we'll create it.
+        cachepath='/Users/robey/code/bzr/bzr.dev/.bzr/loggerhead-cache'
 

=== modified file 'loggerhead/branchview.py'
--- loggerhead/branchview.py	2006-12-20 09:29:42 +0000
+++ loggerhead/branchview.py	2006-12-20 21:17:53 +0000
@@ -38,11 +38,12 @@
 from loggerhead.controllers.download_ui import DownloadUI
 
 
-with_history_lock = util.with_lock('_history_lock')
+with_history_lock = util.with_lock('_history_lock', 'History')
 
 
 class BranchView (object):
-    def __init__(self, name, config):
+    def __init__(self, group_name, name, config):
+        self._group_name = group_name
         self._name = name
         self._config = config
         self.log = logging.getLogger('loggerhead.%s' % (name,))
@@ -64,9 +65,15 @@
     config = property(lambda self: self._config)
     
     name = property(lambda self: self._name)
+
+    group_name = property(lambda self: self._group_name)
     
     friendly_name = property(lambda self: self._config.get('branch_name', self._name))
 
+    description = property(lambda self: self._config.get('description', ''))
+    
+    branch_url = property(lambda self: self._config.get('url', ''))
+
     @turbogears.expose()
     def index(self):
         raise HTTPRedirect(self.url('/changes'))
@@ -99,5 +106,9 @@
             elements = [elements]
         if elements[0].startswith('/'):
             elements[0] = elements[0][1:]
-        return turbogears.url([ '/' + self.name ] + elements, **kw)
+        return turbogears.url([ '/' + self.group_name, self.name ] + elements, **kw)
 
+    def last_updated(self):
+        h = self.get_history()
+        change = h.get_changes([ h.last_revid ])[0]
+        return change.date

=== modified file 'loggerhead/changecache.py'
--- loggerhead/changecache.py	2006-12-20 09:29:42 +0000
+++ loggerhead/changecache.py	2006-12-20 21:17:53 +0000
@@ -35,7 +35,7 @@
 from loggerhead.util import decorator
 
 
-with_lock = util.with_lock('_lock')
+with_lock = util.with_lock('_lock', 'ChangeCache')
 
 
 class ChangeCache (object):
@@ -139,7 +139,8 @@
         jump = 100
         for i in xrange(0, len(work), jump):
             r = work[i:i + jump]
-            self.get_changes(r)
+            # must call into history so we grab the branch lock (otherwise, lock inversion)
+            self.history.get_changes(r)
             if self.closed():
                 return
             count += jump

=== modified file 'loggerhead/controllers/__init__.py'
--- loggerhead/controllers/__init__.py	2006-12-20 09:29:42 +0000
+++ loggerhead/controllers/__init__.py	2006-12-20 21:17:53 +0000
@@ -39,24 +39,43 @@
 log = logging.getLogger("loggerhead.controllers")
 
 
+class Group (object):
+    def __init__(self, name, config):
+        self.name = name
+        self.friendly_name = config.get('name', name)
+        self.description = config.get('description', '')
+        
+        self._views = []
+        for view_name in config.sections:
+            log.debug('Configuring (group %r) branch %r...', name, view_name)
+            view = BranchView(name, view_name, config[view_name])
+            setattr(self, view_name, view)
+            self._views.append(view)
+    
+    views = property(lambda self: self._views)
+
+
 class Root (controllers.RootController):
     def __init__(self):
         global my_config
-        self._views = []
-        for branch_name in my_config.sections:
-            log.debug('Configuring branch %r...', branch_name)
-            view = BranchView(branch_name, my_config[branch_name])
-            setattr(self, branch_name, view)
-            self._views.append(view)
+        self._groups = []
+        for group_name in my_config.sections:
+            group = Group(group_name, my_config[group_name])
+            self._groups.append(group)
+            setattr(self, group_name, group)
         
-    @turbogears.expose()
+    @turbogears.expose(template='loggerhead.templates.browse')
     def index(self):
-        # FIXME - display list of branches
-        raise HTTPRedirect(turbogears.url('/bazaar-dev/changes'))
+        return {
+            'groups': self._groups,
+            'util': util,
+            'title': my_config.get('title', ''),
+        }
 
     def check_rebuild(self):
-        for v in self._views:
-            v.check_rebuild()
+        for g in self._groups:
+            for v in g.views:
+                v.check_rebuild()
 
 
 # singleton:

=== modified file 'loggerhead/controllers/atom_ui.py'
--- loggerhead/controllers/atom_ui.py	2006-12-20 09:29:42 +0000
+++ loggerhead/controllers/atom_ui.py	2006-12-20 21:17:53 +0000
@@ -17,6 +17,7 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
 
+import cherrypy
 import turbogears
 
 from loggerhead import util
@@ -38,7 +39,7 @@
         entries = list(h.get_changes(list(revid_list)[:pagesize]))
 
         vals = {
-            'external_url': self._branch.config.get('external_url'),
+            'external_url': cherrypy.request.base,
             'branch': self._branch,
             'changes': entries,
             'util': util,

=== modified file 'loggerhead/history.py'
--- loggerhead/history.py	2006-12-20 09:29:42 +0000
+++ loggerhead/history.py	2006-12-20 21:17:53 +0000
@@ -54,25 +54,18 @@
 import bzrlib.ui
 
 
-@decorator
-def with_branch_lock(unbound):
-    def branch_locked(self, *args, **kw):
-        self._lock.acquire()
-        try:
-            return unbound(self, *args, **kw)
-        finally:
-            self._lock.release()
-    return branch_locked
-
+with_branch_lock = util.with_lock('_lock', 'branch')
 
 @decorator
 def with_bzrlib_read_lock(unbound):
     def bzrlib_read_locked(self, *args, **kw):
+        #self.log.debug('-> %r bzr lock', id(threading.currentThread()))
         self._branch.repository.lock_read()
         try:
             return unbound(self, *args, **kw)
         finally:
             self._branch.repository.unlock()
+            #self.log.debug('<- %r bzr lock', id(threading.currentThread()))
     return bzrlib_read_locked
 
 
@@ -537,6 +530,7 @@
         cPickle.dump(stats, open('lsprof.stats', 'w'), 2)
         return ret
 
+    @with_branch_lock
     @with_bzrlib_read_lock
     def get_changes_uncached(self, revid_list, get_diffs=False):
         try:

=== modified file 'loggerhead/static/css/style.css'
--- loggerhead/static/css/style.css	2006-12-20 00:43:10 +0000
+++ loggerhead/static/css/style.css	2006-12-20 21:17:53 +0000
@@ -552,3 +552,55 @@
 .annotate th.text {
     padding-left: 1em;
 }
+
+
+/* ----- top-level browse page ----- */
+
+.browse-group-name {
+    background-color: #dee7ec;
+    padding-top: 5px;
+    padding-bottom: 5px;
+    margin-left: 1em;
+    margin-right: 1em;
+    padding-left: 1em;
+    width: 90%;
+    font-weight: bold;
+}
+
+.browse-view {
+	margin-top: 1em;
+	margin-bottom: 1em;
+	margin-left: 2em;
+}
+
+.browse-view table {
+    border-collapse: collapse;
+    width: 90%;
+    margin-bottom: 2em;
+}
+
+.browse-view table tr.heading {
+    xbackground-color: #8cacbb;
+    border-bottom: solid 1px #8cacbb;
+    font-size: smaller;
+}
+
+.browse-view td,.browse-view th {
+	padding: 0.25em 0.5em 0.25em 0.5em;
+}
+
+.browse-view td.name {
+	width: 25%;
+}
+
+.browse-view td.last-update {
+	font-size: smaller;
+	width: 20%;
+}
+
+.browse-view td.url {
+	font-family: monospace;
+	font-size: smaller;
+	padding-left: 3em;
+	padding-top: 0px;
+}

=== added file 'loggerhead/templates/browse.kid'
--- loggerhead/templates/browse.kid	1970-01-01 00:00:00 +0000
+++ loggerhead/templates/browse.kid	2006-12-20 21:17:53 +0000
@@ -0,0 +1,45 @@
+<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
+<html xmlns="http://www.w3.org/1999/xhtml" xmlns:py="http://purl.org/kid/ns#"
+    py:extends="'master.kid'">
+<head>
+    <meta content="text/html; charset=utf-8" http-equiv="Content-Type" py:replace="''"/>
+    <title> loggerhead branches </title>
+</head>
+
+<body>
+
+<h1>
+    <span py:if="title">${title}</span>
+    <span py:if="not title"> bazaar branches in loggerhead </span>
+</h1>
+
+<div class="browse-group" py:for="group in groups">
+    <div class="browse-group-name">
+        ${group.friendly_name}
+    </div>
+    
+    <div class="browse-view">
+        <table>
+        <tr class="heading">
+            <th> Name </th>
+            <th> Description </th>
+            <th> Last change </th>
+        </tr>
+        <span py:for="view in group.views">
+            <tr>
+            <td class="name"> <a href="${tg.url([ '/' + group.name, view.name ])}">${view.friendly_name}</a> </td>
+            <td class="description"> ${view.description} </td>
+            <td class="last-update"> <!--${view.last_updated().strftime('%d %b %Y')} &nbsp;--> ${util.ago(view.last_updated())} </td>
+            </tr>
+            <tr py:if="view.url">
+                <td class="name"> </td>
+                <td class="description url" colspan="2"> <a href="${view.url}">${view.branch_url}</a> </td>
+            </tr>
+        </span>
+        </table>
+    </div>
+
+</div>
+
+</body>
+</html>

=== modified file 'loggerhead/templates/master.kid'
--- loggerhead/templates/master.kid	2006-12-20 09:29:42 +0000
+++ loggerhead/templates/master.kid	2006-12-20 21:17:53 +0000
@@ -13,7 +13,7 @@
 <!-- !define common navbar -->
 <span py:def="navbar()">
     <!-- !requires: ${navigation: start_revid, revid, revid_list, pagesize, buttons, scan_url}, ${branch}, ${history} -->
-    <div class="navbar">
+    <div class="navbar" py:if="navigation is not None">
         <div class="bar">
             <!-- form must go OUTSIDE the table, or safari will add extra padding :( -->
             <form action="${branch.url('/changes', start_revid=getattr(navigation, 'start_revid', None), file_id=getattr(navigation, 'file_id', None))}">

=== modified file 'loggerhead/util.py'
--- loggerhead/util.py	2006-12-20 09:29:42 +0000
+++ loggerhead/util.py	2006-12-20 21:17:53 +0000
@@ -299,15 +299,19 @@
 
 
 # common threading-lock decorator
-def with_lock(lockname):
+def with_lock(lockname, debug_name=None):
+    if debug_name is None:
+        debug_name = lockname
     @decorator
     def _decorator(unbound):
         def locked(self, *args, **kw):
+            #self.log.debug('-> %r lock %r', id(threading.currentThread()), debug_name)
             getattr(self, lockname).acquire()
             try:
                 return unbound(self, *args, **kw)
             finally:
                 getattr(self, lockname).release()
+                #self.log.debug('<- %r unlock %r', id(threading.currentThread()), debug_name)
         return locked
     return _decorator
 

